﻿using System;
using System.Diagnostics;
using System.IO;
using System.Threading;
using Microsoft.DirectX.DirectSound;
using DSBuffer = Microsoft.DirectX.DirectSound.Buffer;
using System.Collections.Generic;

namespace Tscharlii
{
    public class Sound : ISound, IDisposable
    {
        private const string LOG_PREFIX = "Tscharlii::Sound";

        private const string PROCESS_NAME = "sh5";
        //private const string PROCESS_NAME = "TscharliiTest.vshost";

        private const int SHARED_DATA_MUTEX = 0;
        private const int PLAYBACK_FINISHED_EVENT = 1;
        private const int NUM_HANDLES = 2;

        private Device dsDevice = new Device();
        private WaitHandle[] handles = new WaitHandle[NUM_HANDLES];
        private Dictionary<string, DSBuffer> soundDict = new Dictionary<string, DSBuffer>();
        private Queue<DSBuffer> soundQueue = new Queue<DSBuffer>();
        private bool exiting = false;
        private bool disposed = false;

        public Sound()
        {
            IntPtr sh5WindowHandle = getSH5WindowHandle();
            initDirectSound(sh5WindowHandle);
            createHandles();
        }

        ~Sound()
        {
            Dispose(false);
        }

        private IntPtr getSH5WindowHandle()
        {
            Process[] processes = Process.GetProcesses();
            foreach (Process process in processes)
            {
                if (process.ProcessName == PROCESS_NAME)
                {
                    return process.MainWindowHandle;
                }
            }
            throw new SoundException(LOG_PREFIX + "::getSH5WindowHandle: Silent Hunter 5 process not found");
        }

        private void initDirectSound(IntPtr sh5WindowHandle)
        {
            dsDevice.SetCooperativeLevel(sh5WindowHandle, CooperativeLevel.Priority);
        }

        private void disposeDirectSound()
        {
            dsDevice.Dispose();
        }

        private void createHandles()
        {
            // Create the mutex that serializes access to the sound queue 
            handles[SHARED_DATA_MUTEX] = new Mutex();
            // Create the PlaybackFinished event. Autoreset means, only one pending playSoundProc thread will continue.
            // This event is signaled, as soon as a single sound finishes playing, so the next sound in the queue can start playing
            // The event is signaled at the beginning, so the first sound that enters the queue can start playing
            handles[PLAYBACK_FINISHED_EVENT] = new AutoResetEvent(true);
        }

        private Mutex getSharedDataMutex()
        {
            return (Mutex)handles[SHARED_DATA_MUTEX];
        }

        private EventWaitHandle getPlayBackFinishedEvent()
        {
            return (EventWaitHandle)handles[PLAYBACK_FINISHED_EVENT];
        }

        public bool addSound(string soundId, string waveFileName)
        {
            if (! soundDict.ContainsKey(soundId))
            {
                if (File.Exists(waveFileName))
                {
                    try
                    {
                        // We want to be able to control the volume, and get notified, when the playback finished
                        BufferDescription bd = new BufferDescription();
                        bd.GlobalFocus = false;
                        bd.ControlVolume = true;
                        bd.ControlPositionNotify = true;

                        // create the buffer and set the volume
                        DSBuffer buffer = new DSBuffer(waveFileName, bd, dsDevice);
                        buffer.Volume = -1500;

                        // Make the buffer signal the playback finished event, once it stops playing
                        BufferPositionNotify[] positionNotifies = new BufferPositionNotify[1];
                        positionNotifies[0] = new BufferPositionNotify();
                        positionNotifies[0].Offset = (int)PositionNotifyFlag.OffsetStop;
                        positionNotifies[0].EventNotifyHandle = getPlayBackFinishedEvent().SafeWaitHandle.DangerousGetHandle();
                        Notify notify = new Notify(buffer);
                        notify.SetNotificationPositions(positionNotifies);
                        notify.Dispose();

                        // save the buffer in the sound dictionary
                        soundDict.Add(soundId, buffer);
                        return true;
                    }
                    catch (Exception e)
                    {
                        Trace.WriteLine(LOG_PREFIX + "::addSound: Exception: " + e.Message);
                    }
                }
                else
                {
                    Trace.WriteLine(LOG_PREFIX + "::addSound: waveFileName \"" + waveFileName + "\" does not exist");
                }
            }
            else
            {
                Trace.WriteLine(LOG_PREFIX + "::addSound: soundId \"" + soundId + "\" already registered");
            }
            return false;
        }

        public bool playSound(string soundId)
        {
            DSBuffer buffer;
            if (soundDict.TryGetValue(soundId, out buffer))
            {
                // acquire queue mutex
                getSharedDataMutex().WaitOne();

                // Do not accept new sounds when exiting, release the mutex and immediately return
                if (exiting)
                {
                    getSharedDataMutex().ReleaseMutex();
                    return false;
                }

                // enqueue the sound into the sound queue
                soundQueue.Enqueue(buffer);

                // release the queue mutex
                getSharedDataMutex().ReleaseMutex();

                // create a thread that starts playing the sound
                ThreadPool.QueueUserWorkItem(new WaitCallback(playSoundThread));
            }
            else
            {
                Trace.WriteLine(LOG_PREFIX + "playSound: soundId \"" + soundId + "\" does not exist");
            }
            return false;
        }

        public int getNumberOfQueuedSounds()
        {
            // acquire queue mutex
            getSharedDataMutex().WaitOne();

            int numberOfQueuedSounds = soundQueue.Count;

            // release the queue mutex
            getSharedDataMutex().ReleaseMutex();

            return numberOfQueuedSounds;
        }

        private void playSoundThread(Object stateInfo)
        {
            // acquire queue mutex and wait for playback finished event
            WaitHandle.WaitAll(handles);
 
            // Get the next sound to play
            DSBuffer buffer = soundQueue.Dequeue();

            // The destructor signals Sound destruction, so don't play the sound anymore, but signal the next pending thread to continue and exit
            // We removed our sound buffer from the queue before. The destructor can clean up, once the queue is empty, ie. all pending threads finished.
            if (exiting)
            {
                Trace.WriteLine(LOG_PREFIX + "::playSoundThread: Skipping Playback");
                getPlayBackFinishedEvent().Set();
            }
            // Otherwise start playing the sound from the beginning
            else
            {
                buffer.SetCurrentPosition(0);
                buffer.Play(0, BufferPlayFlags.Default);
            }

            // Release the sound queue mutex
            // The playback finished event will be signaled by the buffer, once playback finished
            // Then the next playSoundProc thread can play the next sound in the queue
            getSharedDataMutex().ReleaseMutex();
        }

        private void letAllPendingPlaySoundThreadsFinish()
        {
            // Acquire queue mutex
            getSharedDataMutex().WaitOne();

            // set exiting mode and signal the playback finished event, so pending playSoundThreads continue
            exiting = true;
            getPlayBackFinishedEvent().Set();

            // release queue mutex
            getSharedDataMutex().ReleaseMutex();

            bool allThreadsExited = false;
            while (!allThreadsExited)
            {
                // Acquire queue mutex
                getSharedDataMutex().WaitOne();

                // No pending sounds in the sound queue? Then there are no threads left
                if (soundQueue.Count == 0)
                {
                    allThreadsExited = true;
                }

                // release queue mutex
                getSharedDataMutex().ReleaseMutex();
            }

        }

        private void stopAllSoundsAndClearSoundDict()
        {
            foreach (KeyValuePair<string, DSBuffer> sound in soundDict)
            {
                sound.Value.Stop();
                sound.Value.Dispose();
            }
            soundDict.Clear();
        }

        public void Dispose()
        {
            Dispose(true);

            // Call SupressFinalize in case a subclass implements a finalizer.
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (! disposed)
            {
                if (disposing)
                {
                    letAllPendingPlaySoundThreadsFinish();
                    stopAllSoundsAndClearSoundDict();
                    disposeDirectSound();
                }

                // Indicate that the instance has been disposed.
                disposed = true;
            }
        }
    }
}
